// EJECT - eject devices [quietly]
//
// Copyleft 2014, Volker Birk, Oberer Graben 4, 8400 Winterthur, Switzerland
// mailto:dingens@bumens.org
//
// This program is under GNU General Public License v2

// eject.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

using namespace std;

bool quiet = false;

struct SystemErrorException : public exception {
	DWORD err;
	SystemErrorException(DWORD _err) : err(_err) { }
};

void eject(wstring dev) {
	if (!quiet)
		wcout << L"Ejecting device " << dev << L"\n";

	HANDLE hDevice = CreateFile(dev.c_str(), GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hDevice == INVALID_HANDLE_VALUE)
		throw SystemErrorException(GetLastError());

	DWORD bytesReturned;

	BOOL result = DeviceIoControl(hDevice, IOCTL_STORAGE_EJECT_MEDIA, NULL, 0, NULL, 0, &bytesReturned, NULL);
	if (result == FALSE) {
		DWORD _err = GetLastError();
		CloseHandle(hDevice);
		throw SystemErrorException(_err);
	}

	CloseHandle(hDevice);
}

void showUsage() {
	wcerr << L"EJECT [/Q] <device1> [<device2> [<device3> ...]]\n\n"
		L"Ejects devices [quietly]\n\nDevices can be identified by drive letter (i.e. F:) or device name.\n";
	exit(0);
}

int _tmain(int argc, _TCHAR* argv[])
{
	list<wstring> devices;

	for (int i=1; i<argc; i++) {
		if (argv[i][0] == L'/' || argv[i][0] == L'-') {
			if (((argv[i][1] | (L'a'^L'A')) == L'h') || argv[i][1] == L'?')
				showUsage();
			if ((argv[i][1] | (L'a'^L'A')) == L'q') {
				quiet = true;
				continue;
			}
			wcerr << L"illegal parameter: " << argv[i] << L"\n";
			exit(1);
		}

		wstring device;
		if (argv[i][0] == L'\\')
			device = argv[i];
		else {
			device = L"\\\\.\\";
			device += argv[i];
		}
		devices.push_back(device);
	}

	if (devices.empty())
		showUsage();

	bool err = false;

	for (auto i=devices.begin(); i!=devices.end(); ++i) {
		try {
			eject(*i);
		}
		catch (SystemErrorException& ex) {
			if (!quiet) {
				LPTSTR lpMsgBuf;
				FormatMessage(
						FORMAT_MESSAGE_ALLOCATE_BUFFER | 
						FORMAT_MESSAGE_FROM_SYSTEM |
						FORMAT_MESSAGE_IGNORE_INSERTS,
						NULL,
						ex.err,
						MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
						(LPTSTR) &lpMsgBuf,
						0, NULL
					);

				wcout << L"Cannot eject " << *i << L" - " << lpMsgBuf;
				LocalFree(lpMsgBuf);
			}
			err = true;
		}
	}

	if (err)
		return 2;
	else
		return 0;
}
